
/* s1p.c

   S1 cpu micro architecture simulator

   Computer Architecture 1997

   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

   pipeline    15 Aug 97
*/

#include <stdio.h>
#include "s1p.h"

int R[8], Ir, Mdr, Mar, Pc, T, Z, S, M[1024];
int Rbusy[8];
int Irx, Irm, Irw;
int inst, instx, instm, instw;
int T1, Lmdr, Smdr;
int clock, stall;
int trace, core, start, nbyte;   /* control display */

/* this is not a good style
   but make it simpler to compile */
#include "supportp.c"

/* affect condition code Z,S  */
alu(char op, int a, int b)
{
  switch(op) {
  case '+': T = a+b; break;
  case '-': T = a-b; break;
  case 'c': T = a-b; break;
  }
  if( T == 0 ) Z = 1; else Z = 0;
  if( T < 0 )  S = 1; else S = 0;
}

testCc(int cond )	/* test condition code */
{
  switch( cond ) {    	/*  -,z,nz,lt,le,ge,gt */
  case 0: return 1;	/*  0,1, 2, 3, 4, 5, 6 */
  case 1: return Z;
  case 2: return !Z;
  case 3: return S;
  case 4: return S || Z;
  case 5: return !S || Z;
  case 6: return !S;
  }
  return 0;
}

chkhazard()
{
  switch( instx  ) {
  case ST:   return( isbusy(IRr(Irx)) );
  case MOV:
  case LDR:
  case INC:  return( isbusy(IRr1(Irx)) );
  case STR:
  case ADD:
  case CMP:  return( isbusy(IRr1(Irx)) ||
		     isbusy(IRr2(Irx)) );
  }
}

setwriteR()	/* mark register as busy */
{
  switch( instx ) {
  case LD:   setbusy(IRr(Irx)); break;
  case MOV:
  case LDR:  setbusy(IRr2(Irx)); break;
  case ADD:
  case INC:  setbusy(IRr1(Irx)); break;
  }
}

wstage()
{
  switch( instw ) {
  case LD:  R[IRr(Irw)] = Lmdr; break;
  case MOV: R[IRr2(Irw)] = T1; break;
  case LDR: R[IRr2(Irw)] = Lmdr; break;
  case ADD:
  case INC: R[IRr1(Irw)] = T1; break;
  }
}

mstage()
{
  switch( instm ) {
  case LD:
  case LDR: Lmdr = M[Mar]; break;
  case ST:
  case STR: M[Mar] = Smdr; break;
  case MOV:
  case ADD:
  case INC: T1 = T; break;
  }
}

/* the usual sequence of pipe execution :
   W M X D F -> W M X D F   where -> is movepipe
   when RAW stall
   W M m>w W X D F ...      where m>w is movepipe M to W
   when X is jump taken
   W M X F D F ..	    to fetch two new instructions
*/
xstage()
{
  if(chkhazard()) {   /* stall */
    stall += 2;
    movepipemw();
    wstage();
    clearallbusy();
  }
  setwriteR();
  switch( instx ) {
  case LD:  Mar = IRads(Irx);
	    break;
  case ST:  Mar = IRads(Irx); Smdr = R[IRr(Irx)];
	    break;
  case MOV: T = R[IRr1(Irx)];
	    break;
  case LDR: Mar = R[IRr1(Irx)];
	    break;
  case STR: Mar = R[IRr2(Irx)]; Smdr = R[IRr1(Irx)];
	    break;
  case ADD: alu('+',R[IRr1(Irx)], R[IRr2(Irx)] );
	    break;
  case CMP: alu('c',R[IRr1(Irx)], R[IRr2(Irx)] );
	    break;
  case INC: alu('+',R[IRr1(Irx)], 1);
	    break;
  case JMP: if( testCc(IRr(Irx)) ) {
	      Pc = IRads(Irx);
	      stall += 2;		/* stall */
	      fstage();
	    }
	    break;
  case CALL:  trap(IRads(Irx));
	    break;
  }
}

dstage()
{
  Ir = Mdr;  inst = decodeIR();
}

fstage()
{
  Mdr = M[Pc];  Pc = Pc + 1;
}

movepipe()
{
  instw = instm; instm = instx; instx = inst;
  Irw = Irm; Irm = Irx; Irx = Ir;
}

movepipemw()
{
  instw = instm;
  Irw = Irm;
}

run()
{
  while(1) {
    clock++;
    wstage();
    mstage();
    xstage();
    dstage();
    fstage();
    movepipe();
    if(trace) show();
  }
}

main()
{
  trace = 0;
  core = 0;
  loadprogram();
  stall = 0;
  clock = 0;
  Pc = 0;
  Z = 0;
  S = 0;
  clearallbusy();
  inst = instx = instm = instw = NOP;
  Mdr = 0xff00;
  run();
}
